# -----------------------------------------------------------------------------
#   Copyright (C): OpenGATE Collaboration
#   This software is distributed under the terms
#   of the GNU Lesser General  Public Licence (LGPL)
#   See LICENSE.md for further details
# -----------------------------------------------------------------------------

import numpy as np
import logging
logger=logging.getLogger(__name__)

class bounding_box(object):
    """
    Define ranges in which things are in 3D space.
    Maybe this should be more dimensionally flexible, to also handle 2D or N
    dimensional bounding boxes.
    """
    def __init__(self,**kwargs):
        """
        Initialize using *one* of the following kwargs:
        * 'bb': copy from another bounding_box object
        * 'img': obtain bounding box from a 'itk' image
        * 'xyz': initialize bounding box with 6 floats, shaped (x1,y1,z1,x2,y2,z2),
          ((x1,x2),(y1,y2),(z1,z2)) or ((x1,y1,z1),(x2,y2,z2)).
        TODO: maybe 'extent' would be a better name for the "limits" data member.
        TODO: maybe the different kind of constructors should be implemented as static methods instead of with kwargs.
        """
        nkeys = len(kwargs.keys())
        self.limits=np.empty((3,2))
        if nkeys == 0:
            self.reset()
        elif nkeys > 1:
            raise RuntimeError("too many arguments ({}) to bounding box constructor: {}".format(nkeys,kwargs))
        elif "bb" in kwargs:
            bb = kwargs["bb"]
            self.limits = np.copy(bb.limits)
        elif "img" in kwargs:
            img = kwargs["img"]
            if len(img.GetOrigin()) != 3:
                raise ValueError("only 3D bounding boxes/images are supported")
            origin = np.array(img.GetOrigin())
            spacing = np.array(img.GetSpacing())
            dims = np.array(img.GetLargestPossibleRegion().GetSize())
            self.limits[:,0] = origin-0.5*spacing
            self.limits[:,1] = origin+(dims-0.5)*spacing
        elif "xyz" in kwargs:
            xyz = np.array(kwargs["xyz"],dtype=float)
            if xyz.shape==(3,2):
                self.limits = xyz
            elif xyz.shape==(2,3):
                self.limits = xyz.T
            elif xyz.shape==(6,):
                self.limits = xyz.reshape(3,2)
            else:
                raise ValueError("unsupported shape for xyz limits: {}".format(xyz.shape))
            if np.logical_not(self.limits[:,0]<=self.limits[:,1]).any():
                raise ValueError("min should be less or equal max but I got min={} max={}".format(self.limits[:,0],self.limits[:,1]))
    def reset(self):
        self.limits[:,0]=np.inf
        self.limits[:,1]=-np.inf
    def __repr__(self):
        return "bounding box [[{},{}],[{},{}],[{},{}]]".format(*(self.limits.flat[:].tolist()))
    @property
    def volume(self):
        if np.isinf(self.limits).any():
            return 0.
        return np.prod(np.diff(self.limits,axis=1))
    @property
    def empty(self):
        return (self.volume == 0.)
    def __eq__(self,rhs):
        if self.empty and rhs.empty:
            return True
        return (self.limits==rhs.limits).all()
    def should_contain(self,point):
        apoint = np.array(point,dtype=float)
        assert(len(apoint.shape)==1)
        assert(apoint.shape[0]==3)
        self.limits[:,0] = np.min([self.limits[:,0],apoint],axis=0)
        self.limits[:,1] = np.max([self.limits[:,1],apoint],axis=0)
    def should_contain_all(self,points):
        assert(np.array(points).shape[1]==3)
        self.should_contain(np.min(points,axis=0))
        self.should_contain(np.max(points,axis=0))
    @property
    def mincorner(self):
        return self.limits[:,0]
    @property
    def maxcorner(self):
        return self.limits[:,1]
    @property
    def center(self):
        return 0.5*(self.mincorner+self.maxcorner)
    def contains(self,point,inner=False):
        assert(len(point)==3)
        if inner:
            return ((point>self.limits[:,0])*(point<self.limits[:,1])).all()
        else:
            return ((point>=self.limits[:,0])*(point<=self.limits[:,1])).all()
    def encloses(self,bb,inner=False):
        return (self.contains(bb.mincorner,inner) and self.contains(bb.maxcorner,inner))
    def __contains__(self,item):
        """
        Support for the 'in' operator

        Works only for other bounding boxes, and for 3D points represented by numpy arrays of shape (3,).
        """
        if type(item)==type(self):
            return self.encloses(item)
        else:
            return self.contains(item)
    def add_margins(self,margins):
        # works both with scalar and vector
        # TODO: allow negative margins, but implement appropriate behavior in case margin is larger than the bb.
        self.limits[:,0]-=margins
        self.limits[:,1]+=margins
    def merge(self,bb):
        if self.empty:
            self.limits = np.copy(bb.limits)
        elif not bb.empty:
            self.should_contain(bb.mincorner)
            self.should_contain(bb.maxcorner)
    def have_overlap(self,bb):
        # not sure this is correct!
        return ((not self.empty) and (not bb.empty) and (self.limits[:,0]<=bb.limits[:,1]).all() and (bb.limits[:,0]<=self.limits[:,1]).all())
    def intersect(self,bb):
        if not self.have_overlap(bb):
            self.reset()
        else:
            self.limits[:,0] = np.max([self.mincorner,bb.mincorner],axis=0)
            self.limits[:,1] = np.min([self.maxcorner,bb.maxcorner],axis=0)
    @property
    def xmin(self):
        return self.limits[0,0]
    @property
    def xmax(self):
        return self.limits[0,1]
    @property
    def ymin(self):
        return self.limits[1,0]
    @property
    def ymax(self):
        return self.limits[1,1]
    @property
    def zmin(self):
        return self.limits[2,0]
    @property
    def zmax(self):
        return self.limits[2,1]

#######################################################################
# TESTING
#######################################################################
import unittest

class test_bounding_box(unittest.TestCase):
    def test_xyz_constructor(self):
        bb0 = bounding_box(xyz=[[1,2],[3,4],[5,6]])
        self.assertTrue( ( bb0.limits.flat == np.arange(1,7) ).all() )
        bb1 = bounding_box(xyz=[1,2,3,4,5,6])
        self.assertTrue( ( bb1.limits.flat == np.arange(1,7) ).all() )
        bb1 = bounding_box(xyz=range(1,7))
        self.assertTrue( ( bb1.limits.flat == np.arange(1,7) ).all() )
        with self.assertRaises(ValueError):
            bbxyz_wrong = bounding_box(xyz=[[1,2],[4,3],[5,6]])
        with self.assertRaises(ValueError):
            bbxyz_wrong = bounding_box(xyz=[[np.nan,2],[4,3],[5,6]])
        with self.assertRaises(ValueError):
            bbxyz_wrong = bounding_box(xyz=[[1,2],[3],[5,6,7]])
        with self.assertRaises(ValueError):
            bbxyz_wrong = bounding_box(xyz=[[1,2],[3],[5,6]])
        with self.assertRaises(ValueError):
            bbxyz_wrong = bounding_box(xyz=[1,2,3,4,5,6,7])
        with self.assertRaises(ValueError):
            bbxyz_wrong = bounding_box(xyz=[1,2,3,4,5])
    def test_max_one_key_constructor(self):
        with self.assertRaises(RuntimeError):
            bb0 = bounding_box()
            bbxyz = bounding_box(xyz=[[1,2],[3,4],[5,6]],bb=bb0)
    def test_bb_constructor(self):
        bbxyz = bounding_box(xyz=[[1,2],[3,4],[5,6]])
        bbxyz2 = bounding_box(bb=bbxyz)
        self.assertTrue( (bbxyz.limits == bbxyz2.limits).all() )
        self.assertFalse( bbxyz.limits is bbxyz2.limits ) # it should be a *copy*, not a reference to the same object
    def test_default_constructor(self):
        bb0 = bounding_box()
        self.assertTrue( np.isinf(bb0.limits).all() )
        self.assertTrue( (bb0.limits[:,0]>0).all() )
        self.assertTrue( (bb0.limits[:,1]<0).all() )
    def test_equal(self):
        bb0 = bounding_box(xyz=[[1,2],[3,4],[5,6]])
        bb1 = bounding_box(xyz=[[1,2],[3,4],[5,6]]) # same
        bb2 = bounding_box(xyz=[[1,2],[3,5],[5,6]]) # different
        self.assertTrue(bb0==bb1)
        self.assertFalse(bb0==bb2)
        bb0 = bounding_box()
        bb1 = bounding_box()
        self.assertTrue(bb0==bb1) # empty is empty
        self.assertFalse(bb0==bb2) # empty is not full
        bb0 = bounding_box()
        bb1 = bounding_box(bb=bb0)
        self.assertTrue(bb0==bb1)
        self.assertFalse(bb1==bb2)
        bb1 = bounding_box(xyz=[[1,2],[3,4],[5,5]])
        self.assertTrue(bb0==bb1) # both bounding boxes are empty
        self.assertFalse(bb1==bb2)
    def test_corners(self):
        bbxyz = bounding_box(xyz=[[1,2],[3,4],[5,6]])
        self.assertTrue((bbxyz.mincorner == np.array([1.,3.,5.])).all())
        self.assertTrue((bbxyz.maxcorner == np.array([2.,4.,6.])).all())
    def test_contains(self):
        bbxyz = bounding_box(xyz=[[1,2],[3,4],[5,6]])
        xx,yy,zz = np.meshgrid(np.arange(0.5,2.6,1.0), np.arange(2.5,4.6,1.0), np.arange(4.5,6.6,1.0) )
        for i,point in enumerate(zip(xx.flat,yy.flat,zz.flat)):
            if i==13:
                self.assertTrue(bbxyz.contains(point))
            else:
                self.assertFalse(bbxyz.contains(point))
        self.assertTrue(bbxyz.contains(bbxyz.mincorner,inner=False))
        self.assertFalse(bbxyz.contains(bbxyz.mincorner,inner=True))
        self.assertTrue(bbxyz.contains(bbxyz.maxcorner,inner=False))
        self.assertFalse(bbxyz.contains(bbxyz.maxcorner,inner=True))
    def test_grow(self):
        bbxyz = bounding_box(xyz=[[1,2],[3,4],[5,6]])
        for point in np.array(range(30)).reshape(10,3):
            bbxyz.should_contain(point)
            self.assertTrue(bbxyz.contains(point,inner=False))
        self.assertFalse(bbxyz.contains((0,0,0)))
        self.assertFalse(bbxyz.contains((0,0,30)))
        self.assertFalse(bbxyz.contains((-1,10,10)))
        bbxyz2 = bounding_box(xyz=[[1,2],[3,4],[5,6]])
        bbxyz2.should_contain_all( np.array(range(30)).reshape(10,3) )
        self.assertEqual(bbxyz,bbxyz2)
    def test_grow_margin(self):
        # scalar grow ##################################################
        bbxyz = bounding_box(xyz=[1,2,3,4,5,6])
        bbxyz.add_margins(0.5)
        self.assertTrue( (bbxyz.limits.flat == np.array([0.5,2.5,2.5,4.5,4.5,6.5])).all() )
        # scalar shrink ################################################
        bbxyz = bounding_box(xyz=[1,2,3,4,5,6])
        bbxyz.add_margins(-0.25)
        self.assertTrue( (bbxyz.limits.flat == np.array([1.25,1.75,3.25,3.75,5.25,5.75])).all() )
        # vector grow ##################################################
        bbxyz = bounding_box(xyz=[1,2,3,4,5,6])
        bbxyz.add_margins(np.array([1,3,5]))
        self.assertTrue( (bbxyz.limits[:,0] == 0.).all() )
        bbxyz = bounding_box(xyz=[1,2,3,4,5,6])
        bbxyz.add_margins(np.array([8,6,4]))
        self.assertTrue( (bbxyz.limits[:,1] == 10.).all() )
    def test_merge(self):
        bbxyz0 = bounding_box(xyz=np.arange(1,7))
        bbxyz1 = bounding_box(xyz=np.array([1.1,1.9,3.1,3.9,5.1,5.9]))
        bbxyz2 = bounding_box(xyz=np.arange(1.1,7,1.1))
        bbxyz0.merge(bbxyz1) # should not change anything
        self.assertTrue( (bbxyz0.limits[:,0] == np.arange(1,7,2)).all() )
        bbxyz0.merge(bbxyz2)
        self.assertTrue( (bbxyz0.limits[:,0] == np.arange(1,5.1,2)).all() ) # lower limits unchanged
        self.assertTrue( (bbxyz0.limits[:,1] == np.arange(1.1,7,1.1)[1::2]).all() ) # upper limits changed
    def test_intersect(self):
        bbxyz0 = bounding_box(xyz=np.arange(1,7))
        bbxyz1 = bounding_box(xyz=np.arange(1,7)+0.5)
        bbxyz0.intersect(bbxyz1)
        self.assertTrue( (bbxyz0.limits[:,0] == np.arange(1.5,5.6,2)).all() ) # lower limits changed
        self.assertTrue( (bbxyz0.limits[:,1] == np.arange(2,7,2)).all() ) # upper limits unchanged



# -----------------------------------------------------------------------------
#   Copyright (C): OpenGATE Collaboration
#   This software is distributed under the terms
#   of the GNU Lesser General  Public Licence (LGPL)
#   See LICENSE.md for further details
# -----------------------------------------------------------------------------

"""
This module provides basic gauss filtering for images.
"""

import os
import itk
import numpy as np
import math
import logging
logger=logging.getLogger(__name__)

def gaussFilter(input=None, sigma=None, sigma_mm=None, float=False):

    if input is None:
        logger.error("Set the input")
        sys.exit(1)
    if sigma is None and sigma_mm is None:
        logger.error("Choose between sigma and sigma_mm options")
        sys.exit(1)
    if sigma is not None and sigma_mm is not None:
        logger.error("Choose between sigma and sigma_mm")
        sys.exit(1)

    imageDimension = input.GetImageDimension()
    if sigma_mm is None:
        sigma_mm = [0]*imageDimension
        for i in range(0, imageDimension):
            sigma_mm[i] = sigma[i]*input.GetSpacing()[i]

    tempImageType = itk.Image[itk.F, imageDimension]
    castImageFilter = itk.CastImageFilter[type(input), tempImageType].New()
    castImageFilter.SetInput(input)
    castImageFilter.Update()

    duplicator = itk.ImageDuplicator[tempImageType].New()
    duplicator.SetInputImage(castImageFilter.GetOutput())
    duplicator.Update()
    outputImage = duplicator.GetOutput()
    for i in range(0, imageDimension):
        if sigma_mm[i] >0:
            gaussFilter = itk.RecursiveGaussianImageFilter[tempImageType, tempImageType].New()
            gaussFilter.SetInput(outputImage)
            gaussFilter.SetDirection(i)
            gaussFilter.SetOrder(0)
            gaussFilter.SetNormalizeAcrossScale(False)
            gaussFilter.SetSigma(sigma_mm[i]) # in mm
            gaussFilter.Update()
            outputImage = gaussFilter.GetOutput()
    if not float:
        castImageFilter2 = itk.CastImageFilter[tempImageType, type(input)].New()
        castImageFilter2.SetInput(outputImage)
        castImageFilter2.Update()
        outputImage = castImageFilter2.GetOutput()

    return outputImage

#####################################################################################
import unittest
import sys
from datetime import datetime
import tempfile
import hashlib
import shutil
from .logging_conf import LoggedTestCase

def createImageExample():
    x = np.arange(-10, 10, 1)
    y = np.arange(-12, 15, 1)
    z = np.arange(-13, 10, 1)
    xx, yy, zz = np.meshgrid(x, y, z)
    image = itk.image_from_array(np.int16(xx))
    image.SetOrigin([7, 3.4, -4.6])
    image.SetSpacing([4, 2, 3.6])
    return image

class Test_Image_Gauss(LoggedTestCase):
    def test_image_gauss(self):
        logger.info('Test_Image_Gauss test_image_gauss')
        image = createImageExample()
        transformImage = gaussFilter(input=image, sigma_mm=[0, 2, 0])
        tmpdirpath = tempfile.mkdtemp()
        itk.imwrite(transformImage, os.path.join(tmpdirpath, "testGauss.mha"))
        with open(os.path.join(tmpdirpath, "testGauss.mha"),"rb") as fnew:
            bytesNew = fnew.read()
            new_hash = hashlib.sha256(bytesNew).hexdigest()
            self.assertTrue("c0b6c44826308f94a76192ee6c32ba173f180a627a0e82b8651f899b3545f230" == new_hash)
        shutil.rmtree(tmpdirpath)



# -----------------------------------------------------------------------------
#   Copyright (C): OpenGATE Collaboration
#   This software is distributed under the terms
#   of the GNU Lesser General  Public Licence (LGPL)
#   See LICENSE.md for further details
# -----------------------------------------------------------------------------

"""
This module provides a tool to convert image to dicom
"""

import os
import itk
import numpy as np
import pydicom
from datetime import datetime
import copy
import logging
logger=logging.getLogger(__name__)

def insertTag(dataset, tag, value, type):
    if tag in dataset:
        dataset[tag].value = value
    else:
        dataset[tag] = pydicom.DataElement(tag, type, value)


def convertTagValue(value, VRtype):
    if VRtype in ["AE", "AS" , "AT", "CS", "DA", "DS", "DT", "IS", "LO", "LT", "OB", "OD", "OF", "OW", "PN", "SH", "ST", "TM", "UI", "UN", "UT"]:
        return str(value)
    elif VRtype in ["SL", "SS", "UL", "US"]:
        return int(value)
    elif VRtype in ["FL", "FD"]:
        return float(value)


def writeDicom(input, dicom=None, output="output.dcm", newseriesuid=False, newstudyuid=False, tags=()):

    # Scale the input:
    inputArray = itk.array_from_image(input)
    max = np.amax(inputArray)
    min = np.amin(inputArray)
    scaling = (max - min)/(2**16-1)
    intercept = min
    inputArray = (inputArray - intercept)/scaling
    inputArray = (inputArray).astype(np.uint16)
    inputImageChar = itk.image_from_array(inputArray)

    #Write the input as a new dicom
    imageIO = itk.GDCMImageIO.New()
    writer = itk.ImageSeriesWriter.New(Input=inputImageChar, ImageIO=imageIO)
    writer.SetFileNames([output])
    writer.Update()

    #Open the new dicom and change some dicom tags according dicom input or not
    dsImage = pydicom.dcmread(output, force=True)
    if dicom is not None:
        dsDicom = pydicom.dcmread(dicom, force=True)
        dsOutput = copy.deepcopy(dsDicom)
    else:
        dsOutput = copy.deepcopy(dsImage)

    now = datetime.now()
    insertTag(dsOutput, 0x00080023, now.strftime("%Y%m%d"), 'DA') #Content Date
    insertTag(dsOutput, 0x00080033, now.strftime("%H%M%S") + ".000000", 'TM') #Content Time
    insertTag(dsOutput, 0x00080060, "OT", 'CS') #Modality
    ss = 1.0
    sp = []
    for i in range(2):
        sp += [input.GetSpacing()[1-i]]
    if input.GetImageDimension() == 3:
        ss = input.GetSpacing()[2]
    insertTag(dsOutput, 0x00180050, ss, 'DS') #Slice Thickness
    insertTag(dsOutput, 0x00180088, ss, 'DS') #Spacing Between Slices
    insertTag(dsOutput, 0x00181020, "gatetools", 'LO') #Software Version(s)
    insertTag(dsOutput, 0x00181030, "", 'LO') #Protocol Name
    insertTag(dsOutput, 0x00201041, 0, 'DS') #Slice Location
    origin = []
    for i in input.GetOrigin():
        origin += [i]
    insertTag(dsOutput, 0x00200032, origin, 'DS') #Image Position (Patient)
    orientation = []
    for i in range(2):
        for j in range(input.GetImageDimension()):
            orientation += [input.GetDirection()(i, j)]
    insertTag(dsOutput, 0x00200037, orientation, 'DS') #Image Orientation (Patient)
    insertTag(dsOutput, 0x00280030, sp, 'DS') #Pixel Spacing
    insertTag(dsOutput, 0x00280106, 0, 'US') #Smallest Image Pixel Value
    insertTag(dsOutput, 0x00280107, 65535, 'US') #Largest Image Pixel Value
    insertTag(dsOutput, 0x00281050, (max + min)/2.0, 'DS') #Window Center
    insertTag(dsOutput, 0x00281051, max - min, 'DS') #Window Width
    insertTag(dsOutput, 0x00281052, intercept, 'DS') #Rescale Intercept
    insertTag(dsOutput, 0x00281053, scaling, 'DS') #Rescale Slope
    insertTag(dsOutput, 0x00281055, "auto", 'LO') #Window Center & Width Explanation
    if input.GetImageDimension() == 3:
        frameOffsetVector = [x * ss for x in range(0, input.GetLargestPossibleRegion().GetSize()[2])]
        insertTag(dsOutput, 0x3004000c, frameOffsetVector, 'DS') #Grid Frame Offset Vector

    if dicom is not None:
        dsOutput.PixelData = dsImage.PixelData
        insertTag(dsOutput.file_meta, 0x00020013, dsImage.file_meta[0x0002,0x0013].value, 'SH') #Implementation Version Name
        insertTag(dsOutput.file_meta, 0x00020016, dsImage.file_meta[0x0002,0x0016].value, 'AE') #Implementation Version Name
        insertTag(dsOutput, 0x00280008, dsImage[0x0028,0x0008].value, 'IS') #NumberOfFrames
        insertTag(dsOutput, 0x00280009, dsImage[0x0028,0x0009].value, 'AT') #Frame Increment Pointer
        insertTag(dsOutput, 0x00280010, dsImage[0x0028,0x0010].value, 'US') #Rows
        insertTag(dsOutput, 0x00280011, dsImage[0x0028,0x0011].value, 'US') #Columns
        insertTag(dsOutput, 0x00280100, dsImage[0x0028,0x0100].value, 'US') #Bits Allocated
        insertTag(dsOutput, 0x00280101, dsImage[0x0028,0x0101].value, 'US') #Bits Stored
        insertTag(dsOutput, 0x00280102, dsImage[0x0028,0x0102].value, 'US') #High Bit
        insertTag(dsOutput, 0x00280103, dsImage[0x0028,0x0103].value, 'US') #Pixel Representation
        insertTag(dsOutput, 0x00281054, dsImage[0x0028,0x1054].value, 'LO') #Rescale Type
        #insertTag(dsOutput, 0x52009229, dsImage[0x5200,0x9229].value, 'SQ') #Shared Functional Groups Sequence
        insertTag(dsOutput, 0x52009230, dsImage[0x5200,0x9230].value, 'SQ') #Per-frame Functional Groups Sequence

        if newstudyuid:
            newStudyInstanceUID = pydicom.uid.generate_uid()
            insertTag(dsOutput, 0x0020000d, newStudyInstanceUID, 'UI') #Study Instance UID
            insertTag(dsOutput, 0x00200052, dsImage[0x0020,0x0052].value, 'UI') #Frame of Reference UID
            newseriesuid = True
        if newseriesuid:
            newSeriesInstanceUID = pydicom.uid.generate_uid()
            insertTag(dsOutput, 0x0020000e, newSeriesInstanceUID, 'UI') #Series Instance UID

    # Change tag from command line:
    for tag in tags:
        try:
            VRtype = pydicom.datadict.dictionary_VR(tag[0])
        except:
            print("The tag " + tag[0] + " is not know in public pydicom dictionary")
            continue
        try:
            value = convertTagValue(tag[1], VRtype)
        except:
            print("Cannot find the correct type for " + tag[0] + " with value " + tag[1] + " and VR " + VRtype)
            continue
        insertTag(dsOutput, tag[0], value, VRtype)

    dsOutput.save_as(output)

def printTags(dicomFile):
    return(pydicom.dcmread(dicomFile, force=True))

#####################################################################################
import unittest
import sys
import tempfile
import hashlib
import shutil
import wget
import gatetools as gt
from .logging_conf import LoggedTestCase

def createImageExample():
    x = np.arange(-10, 10, 1)
    y = np.arange(-12, 15, 1)
    z = np.arange(-13, 10, 1)
    xx, yy, zz = np.meshgrid(x, y, z)
    image = itk.image_from_array(np.int16(xx))
    image.SetOrigin([7, 3.4, -4.6])
    image.SetSpacing([4, 2, 3.6])
    return image

class Test_Write_Dicom(LoggedTestCase):
    def test_write_dicom(self):
        logger.info('Test_Write_Dicom test_write_dicom')
        image = createImageExample()
        tmpdirpath = tempfile.mkdtemp()
        itk.imwrite(image, os.path.join(tmpdirpath, "input.mhd"))
        filenameDcm = wget.download("https://gitlab.in2p3.fr/opengate/gatetools_data/-/raw/master/rtdose.dcm?inline=false", out=tmpdirpath, bar=None)
        writeDicom(image, dicom = os.path.join(tmpdirpath, filenameDcm), output = os.path.join(tmpdirpath, "output.dcm"))
        convertedDicom = gt.read_3d_dicom([os.path.join(tmpdirpath, "output.dcm")])
        itk.imwrite(convertedDicom, os.path.join(tmpdirpath, "output.mhd"))
        with open(os.path.join(tmpdirpath, "output.mhd"),"rb") as fnew:
            bytesNew = fnew.read()
            new_hash = hashlib.sha256(bytesNew).hexdigest()
            self.assertTrue("65fc74e082eb4a69b60c289435564488956ba73bb274637fe0969c2152176308" == new_hash)
        shutil.rmtree(tmpdirpath)



import setuptools

with open("readme.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="gatetools",
    version="0.10.1",
    author="OpenGate collaboration",
    author_email="david.sarrut@creatis.insa-lyon.fr",
    description="Python tools for GATE, see https://github.com/OpenGATE/Gate",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/OpenGATE/GateTools",
    package_dir={ 'gatetools': 'gatetools',
                  'gatetools.phsp': 'gatetools/phsp'},
    packages=['gatetools', 'gatetools.phsp'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        'matplotlib',
        'click',
        'numpy',
        'scipy',
        'pydicom',
        'tqdm',
        'colored',
        'itk>=5.1.0',
        'uproot',
        'wget',
        'python-box'
    ],
    scripts=[
        'bin/gt_gate_info',
        
        'bin/gt_image_uncertainty',
        'bin/gt_image_arithm',
        'bin/gt_image_crop',
        'bin/gt_image_convert',
        'bin/gt_image_statistics',
        'bin/gt_gamma_index',
        'bin/gt_affine_transform',
        'bin/gt_write_dicom',
        'bin/gt_dicom_info',
        'bin/gt_image_gauss',
        'bin/gt_image_resize',
        'bin/gt_dvh',
        'bin/gt_merge_root',
        'bin/gt_morpho_math',

        'bin/gt_dicom_rt_struct_to_image',
        'bin/gt_dicom_rt_pbs2gate',
        
        'bin/gt_phsp_info',
        'bin/gt_phsp_convert',
        'bin/gt_phsp_merge',
        'bin/gt_phsp_plot',
        'bin/gt_phsp_peaks',
    ]
)


# -----------------------------------------------------------------------------
# Uploading the package on pypi

# Steps
# 1 - change version in setup.py file
# 2 - commit, tag. git push --tags
# 3 - setup: python3 setup.py sdist bdist_wheel
# 4 - twine: see below

# On TEST pypi: 
# twine upload --repository-url https://test.pypi.org/legacy/ dist/*
# On REAL pyip
# twine upload  dist/*

# test with
# pip uninstall gatetools
# pip3 install --extra-index-url https://test.pypi.org/simple/ gatetools
# https://test.pypi.org/project/gatetools/



# -----------------------------------------------------------------------------
#   Copyright (C): OpenGATE Collaboration
#   This software is distributed under the terms
#   of the GNU Lesser General  Public Licence (LGPL)
#   See LICENSE.md for further details
# -----------------------------------------------------------------------------

import numpy as np
import os
import tokenize
from io import BytesIO
from matplotlib import pyplot as plt
import uproot
import logging

logger = logging.getLogger(__name__)


# -----------------------------------------------------------------------------
def load(filename, nmax=-1, shuffle=False):
    """ 
    Load a PHSP (Phase-Space) file
    Output is numpy structured array
    """

    b, extension = os.path.splitext(filename)
    nmax = int(nmax)

    if extension == '.root':
        if shuffle:
            logger.error('cannot shuffle on root file for the moment')
            exit(0)
        return load_root(filename, nmax)

    # if extension == '.raw':
    #     return load_raw(filename)

    if extension == '.npy':
        return load_npy(filename, nmax, shuffle)

    logger.error('dont know how to open phsp with extension ',
                 extension,
                 ' (known extensions: .root .npy)')
    exit(0)


# -----------------------------------------------------------------------------
def load_root(filename, nmax=-1):
    """
    Load a PHSP (Phase-Space) file in root format
    Output is numpy structured array
    """

    nmax = int(nmax)
    # Check if file exist
    if not os.path.isfile(filename):
        logger.error("File '" + filename + "' does not exist.")
        exit()

    # Check if this is a root file
    try:
        with uproot.open(filename) as f:
            k = f.keys()
            try:
                psf = f['PhaseSpace']
            except Exception:
                logger.error("This root file does not look like a PhaseSpace, keys are: ",
                              f.keys(), ' while expecting "PhaseSpace"')
                exit()

            # Get keys
            names = [k for k in psf.keys()]
            n = psf.num_entries

            # Convert to arrays (this take times)
            if nmax != -1:
                a = psf.arrays(entry_stop=nmax, library="numpy")
            else:
                a = psf.arrays(library="numpy")

            # Concat arrays
            d = np.column_stack([a[k] for k in psf.keys()])
            # d = np.float64(d) # slow
    except Exception:
        logger.error("File '" + filename + "' cannot be opened, not root file ?")
        exit()

    return d, names, n


# -----------------------------------------------------------------------------
def load_npy(filename, nmax=-1, shuffle=False):
    """
    Load a PHSP (Phase-Space) file in npy
    Output is numpy structured array
    """

    # Check if file exist
    if not os.path.isfile(filename):
        logger.error("File '" + filename + "' does not exist.")
        exit()

    x = np.load(filename, mmap_mode='r')
    n = len(x)
    if nmax > 0:
        if shuffle:
            x = np.random.choice(x, nmax, replace=False)
        else:
            x = x[:nmax]

    data = x.view(np.float32).reshape(x.shape + (-1,))
    # data = np.float64(data) # slow
    return data, list(x.dtype.names), n


# -----------------------------------------------------------------------------
def humansize(nbytes):
    """
    https://stackoverflow.com/questions/14996453/python-libraries-to-calculate-human-readable-filesize-from-bytes
    """

    suffixes = ['B', 'KB', 'MB', 'GB', 'TB', 'PB']
    i = 0
    while nbytes >= 1024 and i < len(suffixes) - 1:
        nbytes /= 1024.
        i += 1
    f = ('%.2f' % nbytes).rstrip('0').rstrip('.')
    return '%s %s' % (f, suffixes[i])


# -----------------------------------------------------------------------------
def save_npy(filename, data, keys):
    """
    Write a PHSP (Phase-Space) file in npy
    """

    dtype = []
    for k in keys:
        dtype.append((k, 'f4'))

    r = np.zeros(len(data), dtype=dtype)
    i = 0
    for k in keys:
        r[k] = data[:, i]
        i = i + 1

    np.save(filename, r)


# -----------------------------------------------------------------------------
def remove_keys(data, keys, rm_keys):
    """
    Remove som keys
    """

    cols = np.arange(len(keys))
    index = []
    if len(rm_keys) == 0:
        return data, keys

    for k in rm_keys:
        if k not in keys:
            logger.error('Error the key', k, 'does not exist in', keys)
            exit(0)
        i = keys.index(k)
        cols = np.delete(cols, i)
        index.append(i)
        for c in index:
            keys.pop(c)
    data = data[:, cols]
    return data, keys


# -----------------------------------------------------------------------------
def str_keys_to_array_keys(keys):
    """
    Convert string of keys to arrays of key
    """

    if keys is None:
        return []
    dd = tokenize.tokenize(BytesIO(keys.encode('utf-8')).readline)
    keys = []
    for toknum, tokval, _, _, _ in dd:
        if tokval != 'utf-8' and tokval != '' and tokval != None:
            keys.append(tokval)
    return keys


# -----------------------------------------------------------------------------
def get_E(data, keys):
    """
    Retrieve the E from a dataset
    """

    try:
        Ei = keys.index('Ekine')
    except:
        try:
            Ei = keys.index('E')
        except:
            raise RuntimeError("Error, cannot find key 'Ekine' nor 'E'. Keys are: ", keys)
    E = data[:, Ei]
    return E, Ei


# -----------------------------------------------------------------------------
def fig_get_sub_fig(ax, i):
    """
    Retrieve a fig nb
    """

    # check if single fig
    if not type(ax) is np.ndarray:
        return ax

    # check if single row/line
    if ax.ndim == 1:
        return ax[i]

    # other cases
    index = np.unravel_index(i, ax.shape)
    return ax[index[0]][index[1]]


# -----------------------------------------------------------------------------
def fig_get_nb_row_col(nfig):
    """
    Compute a fig with adapted row/col for n fig
    """

    nrow = int(np.sqrt(nfig))
    ncol = int(nfig / nrow)
    if ncol * nrow < nfig:
        nrow += 1
    return nrow, ncol


# -----------------------------------------------------------------------------
def fig_rm_empty_plot(nfig, ax):
    """
    Remove empty plot
    """

    nrow, ncol = fig_get_nb_row_col(nfig)
    r = nrow - 1
    i = nfig
    while i < ncol * nrow:
        c = i - int(i / ncol) * ncol
        ax[r, c].set_axis_off()
        i = i + 1

    # -----------------------------------------------------------------------------


def keys_toggle_angle(keys):
    """
    In the list of keys, toggle angleXY to XY or XY to angleXY
    """

    k = keys.copy()
    if 'X' in keys and 'Y' in keys:
        k.remove('X')
        k.remove('Y')
        k.append('angleXY')
    else:
        if 'angleXY' in keys:
            k.append('X')
            k.append('Y')
            k.remove('angleXY')
    return k


# -----------------------------------------------------------------------------
def select_keys(data, input_keys, output_keys):
    """
    Keep only the given keys
    """

    if len(output_keys) == 0:
        logger.error('Error, select_keys is void')
        exit(0)

    cols = []
    for k in output_keys:
        try:
            i = input_keys.index(k)
            cols.append(i)
        except:
            logger.error('Error, cannot find', k, 'in keys:', input_keys)
            exit(0)

    data = data[:, cols]
    return data


# -----------------------------------------------------------------------------
def add_angle(data, keys, k1='X', k2='Y'):
    """
    Add and compute angleXY in the list of keys
    angle = atan2(k2,k1)
    """

    if 'angleXY' in keys:
        return data, keys

    i1 = keys.index(k1)
    i2 = keys.index(k2)
    angle = np.arctan2(data[:, i2], data[:, i1])
    data = np.column_stack((data, angle))
    k = keys.copy()
    k.append('angleXY')
    return data, k


# -----------------------------------------------------------------------------
def add_vector_angle(data, keys, radius, k='angleXY', k1='X', k2='Y'):
    """
    Add X and Y from angleXY
    """

    # nothing to do if already exist
    if k1 in keys and k2 in keys:
        return data, keys

    if k not in keys:
        logger.warning('Cannot convert angle, the key angleXY does not exist in ', keys)

    i = keys.index(k)
    angle = data[:, i]
    dx = radius * np.cos(angle)
    dy = radius * np.sin(angle)
    data = np.column_stack((data, dx))
    data = np.column_stack((data, dy))

    kk = keys.copy()
    kk.append(k1)
    kk.append(k2)
    return data, kk


# -----------------------------------------------------------------------------
def add_missing_angle(data, input_keys, output_keys, radius):
    """
    Add missing keys (angleXY or X+Y)
    """

    if 'angleXY' in output_keys:
        data, input_keys = add_angle(data, input_keys)

    if 'X' in output_keys and 'Y' in output_keys:
        data, input_keys = add_vector_angle(data, input_keys, radius)

    return data, input_keys


# -----------------------------------------------------------------------------
def fig_histo2D(ax, data, keys, k, nbins, color='g'):
    """
    Fig 2D histo
    """

    i1 = keys.index(k[0])
    x = data[:, i1]
    i2 = keys.index(k[1])
    y = data[:, i2]
    if color == 'g':
        cmap = plt.cm.Greens
    if color == 'r':
        cmap = plt.cm.Reds
    if color == 'b':
        cmap = plt.cm.Blues

    counts, xedges, yedges, im = ax.hist2d(x, y, bins=(nbins, nbins), alpha=1, cmap=cmap)
    # , norm=LogNorm())
    plt.colorbar(im, ax=ax)
    ax.set_xlabel(k[0])
    ax.set_ylabel(k[1])


#####################################################################################
import unittest
import hashlib
import wget
import tempfile
import shutil
from gatetools.logging_conf import LoggedTestCase


class Test_Phsp(LoggedTestCase):
    def test_phsp_convert(self):
        logger.info('Test_Phsp test_phsp_convert')
        tmpdirpath = tempfile.mkdtemp()
        filenameRoot = wget.download(
            "https://gitlab.in2p3.fr/opengate/gatetools_data/-/raw/master/phsp.root?inline=false", out=tmpdirpath,
            bar=None)
        data, read_keys, m = load(os.path.join(tmpdirpath, filenameRoot), -1)
        save_npy(os.path.join(tmpdirpath, "testphsp.npy"), data, read_keys)
        with open(os.path.join(tmpdirpath, "testphsp.npy"), "rb") as fnew:
            bytesNew = fnew.read()
            new_hash = hashlib.sha256(bytesNew).hexdigest()
            self.assertTrue("cec796ec5764d039b02e15d504e80ccf2b2c35e0e5380985245262faf0ff0892" == new_hash)
        dataNPY, read_keysNPY, mNPY = load(os.path.join(tmpdirpath, "testphsp.npy"), -1)
        self.assertTrue(np.allclose(131.69868, np.amax(dataNPY[:, 2])))
        shutil.rmtree(tmpdirpath)

    def test_phsp_info(self):
        logger.info('Test_Phsp test_phsp_info')
        tmpdirpath = tempfile.mkdtemp()
        filenameRoot = wget.download(
            "https://gitlab.in2p3.fr/opengate/gatetools_data/-/raw/master/phsp.root?inline=false", out=tmpdirpath,
            bar=None)
        data, read_keys, m = load(os.path.join(tmpdirpath, filenameRoot), -1)
        self.assertTrue("17.27 MB" == humansize(os.stat(os.path.join(tmpdirpath, filenameRoot)).st_size))
        self.assertTrue(np.float32 == data.dtype)
        self.assertTrue(782127 == m)
        self.assertTrue(7 == len(read_keys))
        self.assertTrue(np.allclose(131.69868, np.amax(data[:, 2])))
        shutil.rmtree(tmpdirpath)



# -----------------------------------------------------------------------------
#   Copyright (C): OpenGATE Collaboration
#   This software is distributed under the terms
#   of the GNU Lesser General  Public Licence (LGPL)
#   See LICENSE.md for further details
# -----------------------------------------------------------------------------

import shutil
import platform
import subprocess
import pkg_resources
import os
import itk
import numpy as np

# -----------------------------------------------------------------------------
def print_gate_info(short_verbose, cmd_gate_name = 'Gate'):
    '''
    Print information about 'Gate' executablea and system
    '''

    # get Gate path
    path = shutil.which(cmd_gate_name)

    # get version info
    if path == None:
        path = cmd_gate_name+' not found'
        output = ''
    else:
        result = subprocess.run(['Gate', '--version'], stdout=subprocess.PIPE)
        output = result.stdout.decode("utf-8").splitlines()[0]

    if short_verbose:
        print(path)
        return

    # get gatetools version
    gtv = pkg_resources.get_distribution("gatetools").version

    # get G4 env variables
    g4_list = ['G4ABLADATA', 'G4ENSDFSTATEDATA', 'G4INCLDATA', 'G4LEDATA', 'G4LEVELGAMMADATA', 'G4NEUTRONHPDATA', 'G4PARTICLEXSDATA', 'G4PIIDATA', 'G4RADIOACTIVEDATA', 'G4REALSURFACEDATA', 'G4SAIDXSDATA']

    print('Gate path:         ', path)
    print('Version:           ', output)
    print('Machine type:      ', platform.machine())
    print('Hostname:          ', platform.node())
    print('Platform:          ', platform.platform())
    print('Processor:         ', platform.processor())
    print('Python:            ', platform.python_version())
    print('System:            ', platform.system())
    print('GateTools version: ', gtv)
    for g in g4_list:
        try:
            print('{:<19} {}'.format(g, os.environ[g]))
        except:
            print('{:<19} NOT FOUND'.format(g))


def img_size(image):
    return np.array(image.GetLargestPossibleRegion().GetSize())

def img_spacing(image):
    return np.array(image.GetSpacing())
